#########################################################################################
#                                                                                       #
#                    RasPI BASED TWITTER ENABLED AUTOMATIC DICE ROLLER                  #
#                               Author:  Francesco Zuddas                               #
#                       https://uk.linkedin.com/in/francescozuddas                      #
#                               Version 1.0 - 29 Sep 2016                               #
#                                                                                       #
#    The present project relies on:                                                     #
#               Raspberry PI 3 with Raspian Jessie                                      #
#               PI Camera with focusing 115 degree FOV lens                             #
#               Modified automatic dice roller                                          #
#               Relay HAT - SalesLa 5V 4 Channels Relay Board Module                    #
#               Recommended:                                                            #
#                       Raspberry case (SmartiPi)                                       #
#                       Camera tripod + GoPRO mount                                     #
#                       USB reading light                                               #
#               Software:                                                               #
#                       Python 2.7.12                                                   #
#                       OpenCV                                                          #
#                       Twython 3.4.0                                                   #
#                                                                                       #
#   To work with this you'll need a twitter account and twitter app credentials.        #
#   I got all the hardware parts on Amazon.                                             #
#   Once everything is set up you can operate the device with two twitter commands:     #
#                                                                                       #
#                              @youraccount roll the dice!                              #
#                                                                                       #
#   You'll get a picture with the score. The RasPI will activate the relay shield       # 
#   starting the dice roller for 1 second, then it will switch the relay off.           #
#   It will then wait for two seconds for the dice to stop, and then it will take       #
#   a picture, process it and update the stats (file and graph).                        #
#   The processed picture will be tweeted back to the user.                             #
#                                                                                       #
#                              @youraccount stats please!                               #
#                                                                                       #       
#   You'll get a graph with the stats. The updated stats graph will be tweeted back     #
#   to the user.                                                                        #
#                                                                                       #
#   If Twitter imposed limits are not respected the program sleeps untile the next      #
#   limit reset.                                                                        #
#                                                                                       #
#         This program is free software: you can redistribute it and/or modify          #
#         it under the terms of the GNU General Public License as published by          #
#         the Free Software Foundation, either version 3 of the License, or             #
#         (at your option) any later version.                                           #
#                                                                                       #
#         This program is distributed in the hope that it will be useful,               #
#         but WITHOUT ANY WARRANTY; without even the implied warranty of                #
#         MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
#         GNU General Public License for more details.                                  #
#                                                                                       #
#         You should have received a copy of the GNU General Public License             #
#         along with this program.  If not, see <http://www.gnu.org/licenses/>.         #
#                                                                                       #
#########################################################################################

import sys
import csv
import os
import time
import datetime
import cv2
import numpy
import matplotlib.pyplot as plt
from time import sleep
from twython import Twython, TwythonStreamer, TwythonError, TwythonRateLimitError
from picamera import PiCamera

print (time.strftime("%d/%m/%Y")+" - "+time.strftime("%H:%M:%S"))

#-----------------I/O PIN SETTINGS-------------------
import RPi.GPIO as GPIO
#GPIO.setmode(GPIO.BCM) # BCM numbering
GPIO.setmode(GPIO.BOARD) #Physical numbering. The pin numbers follow the pin numbers on header P1
relay01 = 31 # board PIN 31 (BCM pin 6)
relay02 = 33 # board PIN 33 (BCM pin 13)
relay03 = 35 # board PIN 35 (BCM pin 19)
relay04 = 37 # board PIN 37 (BCM pin 26)
GPIO.setup(relay01, GPIO.OUT) #Setting pin mode
GPIO.setup(relay02, GPIO.OUT) #Setting pin mode
GPIO.setup(relay03, GPIO.OUT) #Setting pin mode
GPIO.setup(relay04, GPIO.OUT) #Setting pin mode

def switchRelay(relayNumber, state): #relay is OFF when pin is HIGH! 
        if state == "OFF":
                GPIO.output(relayNumber, GPIO.HIGH)
        elif state == "ON":
                GPIO.output(relayNumber, GPIO.LOW)
        else:
                print("Unexpected command!")

switchRelay(relay01, "OFF")
switchRelay(relay02, "OFF")
switchRelay(relay03, "OFF")
switchRelay(relay04, "OFF")

#-----------------CAMERA SETTINGS-------------------
camera = PiCamera()
camera.resolution = (960,540)
#camera.rotation = 180
#camera.start_preview()
camera.brightness = 30 #You can alter the brightness setting, which can be set from 0 to 100. The default is 50
camera.contrast = 90 #You can alter the brightness setting, which can be set from 0 to 100. The default is 50

#-----------------TWITTER SETTINGS------------------- 
apiKey = "xxx"
apiSecret = "xxx"
accessToken = "xxx"
accessTokenSecret = "xxx"

userCommand = "@yourTwitterAccountNameHere"

#-----------------INITIALIZE DATABASE FILE------------------- 
if os.path.exists("stats.csv") == False:
        print("Inizializing database...") 
        fZero = open("stats.csv", "wt")
        writer = csv.writer(fZero) #,lineterminator="\n") #use lineterminator="\n" on windows+python3
        row=["Score", "Frequency"]
        writer.writerow(row)
        for i in range (3,19):
                row[0] = str(i)
                row[1] = "0"
                writer.writerow(row)
        print("Done!") 
        fZero.close()
else:
        print("File already exist!")
                

#-----------------DICE ROLLER FUNCTION-------------------
def diceroller():
        switchRelay(relay01, "ON") #Start rolling the dice
        time.sleep(1)
        switchRelay(relay01, "OFF") #Stop after one second
        time.sleep(2)
        camera.capture("dice.jpg") #Wait two seconds for the dice to stop and take a picture
        myImage = cv2.imread("dice.jpg", cv2.IMREAD_GRAYSCALE) # Read image and convert in grayscale (SimpleBlobDetector only works on grayscale images)
#        detector = cv2.SimpleBlobDetector() # Default values of parameters are tuned to extract dark circular blobs
        params = cv2.SimpleBlobDetector_Params() # Setup SimpleBlobDetector parameters
        params.minThreshold = 10;
        params.maxThreshold = 200;
        params.filterByColor = True
        params.blobColor = 0
        params.filterByArea = True 
        params.minArea = 50
        params.filterByCircularity = True
        params.minCircularity = 0.80
        params.filterByConvexity = True 
        params.minConvexity = 0.70
        params.filterByInertia = True
        params.minInertiaRatio = 0.60
        detector = cv2.SimpleBlobDetector(params) # Create a detector with the parameters
        diceSpots = detector.detect(myImage) # Detect dice spots
        myImageEdit = cv2.cvtColor(myImage, cv2.COLOR_GRAY2BGR) #Convert image back to color to add colored text and circles
        for diceSpot in diceSpots: 
                X = int(diceSpot.pt[0]) #Dice spot X coordinate
                Y = int(diceSpot.pt[1]) #Dice spot Y coordinate
                S = int(diceSpot.size) #Dice spot size
                #print S #UNCOMMENT FOR DEBUG
                cv2.circle(myImageEdit, (X,Y), S, (0,0,255), -1, 8, 0) #add red circle of size S to the dice spot coordinates X,Y        
        #myImageEdit = cv2.drawKeypoints(myImage, diceSpots, numpy.array([]), (255,255,0), cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS) # Draw detected blobs as red circles.
        font = cv2.FONT_HERSHEY_SIMPLEX
        cv2.putText(myImageEdit,"Points = "+str(len(diceSpots)),(290,50), font, 2,(0, 255, 0),5) #Write the score on the image
        cv2.imwrite("dice.jpg", myImageEdit) #Save the modified image (red dots + score text)
        return str(len(diceSpots)) #Return the score 

#-----------------CREATE STATISTICS-------------------
def updateStats(score):
        fin = open("stats.csv", "rt") #File containing the statistics
        fout = open("aux.csv", "wt") #Create empty auxiliary file

        reader = csv.reader(fin) #handle to statistic file
        writer = csv.writer(fout) #handle to auxiliary file

        i=0 #reset counter
        x=[0]*16 #reset score array
        y=[0]*16 #reset frequency array
        totalRolls = 0 #reset total rolls counter

        try:
                for row in reader: #for every row in stats.csv...
                        if (row[0].isdigit() and (row[0] == score)): #Process stats.csv: if the first element of the row is a digit, and its equal to the score just made...
                                print ("Score = " + str(score)) #Print score just made
                                print ("Old freq = " + row[1]) #Print frequency for that score reading it from the file
                                new_freq = str(int(row[1]) + 1) #Frequency is now increased by 1
                                print ("New freq = " + new_freq) #Print new frequency value
                                row = [str(score), new_freq] #Overwrite the frequency value in the row array
                                #print f #UNCOMMENT FOR DEBUG
                        writer.writerow(row) #Write the row array to aux.csv
                        if (row[0].isdigit()): #If I'm reading a numeric row (so not the header with the titles)...
                                x[i] = int(row[0]) #Create a numeric array for the score
                                y[i] = int(row[1]) #Create a numeric array for the frequency
                                totalRolls = totalRolls + y[i] #Count the total number of rolls
                                i = i + 1 #Update counter
        finally:
                print("Writing done!")
                #print(x) #UNCOMMENT FOR DEBUG
                #print(y) #UNCOMMENT FOR DEBUG
                fin.close()
                fout.close()
        os.remove("stats.csv") #Remove the old file...
        os.rename("aux.csv", "stats.csv") #...and replace it with aux.csv, that now contains the updated values

       # print open("stats.csv", "rt").read() #UNCOMMENT FOR DEBUG

        #create graph--------------------------------------------------

        plt.bar(x, y, 0.35, align="center", color ="r") #Create bar graph
        plt.xlabel("Score") #Score on the X axis
        plt.ylabel("Frequency") #Frequency on the Y axis
        plt.title(r"Three dice rolling statistics, total rolls =" + str(totalRolls)) #Total rolls since starting of the program printed in title

        plt.xlim(2.5, 18.5) #Adjust X axis limit to center the graph
        plt.xticks(x) #Use scores as X axis labels (with three dice this range from 3 to 18)
        plt.tight_layout()
        plt.grid(True)
        #plt.show() #UNCOMMENT FOR DEBUG - note: showing the plot wont allow to save the plot and viceversa!
        plt.savefig("stats.png")

#-----------------COUNTDOWN FUNCTION-------------------
#note: this function will work properly if the program is launched from the shell. In IDLE instead you will get a "scrolling" countdown...
def countdown(t):
    while t:
        mins, secs = divmod(t, 60)
        timeformat = '{:02d}:{:02d}'.format(mins, secs)
        sys.stdout.write("%s\r" % timeformat)
        sys.stdout.flush()
        time.sleep(1)
        t -= 1
    print("Time elapsed!")

#-----------------STREAM CALLBACK: INSTRUCTION PERFORMED-------------------
class myStreamCallback(TwythonStreamer):
        def on_success(self, tweet):
                if "text" in tweet:
                        twitter = Twython(apiKey,apiSecret,accessToken,accessTokenSecret) #Create a Twython instance with your keys
                        username = "@" + (tweet["user"]["screen_name"]) #Get tweet sender username
                        if (("roll the dice!" in tweet["text"]) and ("stats please!" in tweet["text"])):
                                userScore = diceroller() #Roll dices
                                my_photo = open("dice.jpg","rb")
                                response = twitter.upload_media(media = my_photo)
                                twitter.update_status(status= "Dice anyone? :) "+username+" your score is:", media_ids=[response["media_id"]])
                                print ("==========================================================")
                                print(username+" scored "+userScore+" points. Tweet sent!")
                                updateStats(userScore)
                                my_photo = open("stats.png","rb")
                                response = twitter.upload_media(media = my_photo)
                                twitter.update_status(status= "Latest statistics for you "+username+"!", media_ids=[response["media_id"]])
                                print("Statistics sent to "+username)
                        elif "roll the dice!" in tweet["text"]:
                                userScore = diceroller() #Roll dices
                                my_photo = open("dice.jpg","rb")
                                response = twitter.upload_media(media = my_photo)
                                twitter.update_status(status= "Dice anyone? :) "+username+" your score is:", media_ids=[response["media_id"]])
                                print ("==========================================================")
                                print(username+" scored "+userScore+" points. Tweet sent!")
                                updateStats(userScore)
                        elif "stats please!" in tweet["text"]:
                                my_photo = open("stats.png","rb")
                                response = twitter.upload_media(media = my_photo)
                                twitter.update_status(status= "Latest statistics for you "+username+"!", media_ids=[response["media_id"]])
                                print ("==========================================================")
                                print("Statistics sent to "+username)
                        else:
                                print ("==========================================================")                                
                                print ("Tweet from @%s Date: %s" % (tweet["user"]["screen_name"].encode("utf-8"),tweet["created_at"]))
                                print (tweet["text"].encode("utf-8"))
                        twitter.get_home_timeline() #API call, just to allow us to check the limits
                        xrl_limit = twitter.get_lastfunction_header("x-rate-limit-limit") #This function must be called after an API call
                        xrl_remaining = twitter.get_lastfunction_header("x-rate-limit-remaining") #This function must be called after an API call
                        xrl_reset = int(float(twitter.get_lastfunction_header("x-rate-limit-reset")) - time.time()) #This function must be called after an API call
                        print ("x-rate-limit-limit = "+xrl_limit) #Total available API calls
                        print ("x-rate-limit-remaining = "+xrl_remaining)  #Remaining API calls
                        print ("x-rate-limit-reset = "+str(xrl_reset)) #Time to next reset
                        if (int(xrl_remaining) <= 1):
                                print("Data limit exceeded, sleeping for "+str(xrl_reset)+" seconds...")
                                twitter.update_status(status= "Data limit exceeded, sleeping for "+str(xrl_reset)+" seconds...")
                                self.disconnect()
                                countdown(xrl_reset)        
                                time.sleep(3)
        def on_error(self, status_code, data):
                print ("=== E R R O R ===")
                print (time.strftime("%d/%m/%Y")+" - "+time.strftime("%H:%M:%S"))
                print ("Error code : "+str(status_code))
                self.disconnect() #umcomment to disconnect on errors
                print ("Retrying in 15 minutes...")
                countdown(60*15)
        def on_timeout():
                print ("=== T I M E O U T ===")
                print (time.strftime("%d/%m/%Y")+" - "+time.strftime("%H:%M:%S"))
                self.disconnect() #umcomment to disconnect on errors               


#-----------------CREATE TWYTHON STREAMER-------------------
while True:
        try:
                myStream = myStreamCallback(apiKey, apiSecret, accessToken, accessTokenSecret)
                myStream.statuses.filter(track = userCommand)
        except KeyboardInterrupt:
                GPIO.cleanup() # cleanup all GPIO
                print ("That's all folks!")
                break
        except:
                continue
